#!/usr/bin/perl -w
#
# reverseSTR.pl v1.0
# Last modified January 26, 2005 16:06
# (c) Olaf R.P. Bininda-Emonds
#
# Input:
#   1) A list of taxonomic equivalents for a given data set, listed as taxon <tab> index taxon (as produced by STRindexer.pl)
#	2) A nexus-formatted treefile on which to add the equivalent taxa
#
# Output:
#   A nexus-formatted treefile with equivalent taxa added in where possible (and identified by being on zero-length branches)
#
# Usage: reverseSTR.pl -r<STR file> -r<treefile> [-v] [-h]
#   options: -r<filename> = STRindexer.pl file listing taxonomic equivalents
#			 -t<treefile> = tree file containing gene trees in NEXUS format
#			 -v = verbose output
#            -h = print this message and quit

use strict;

# Set user-input defaults and associated parameters
	# Supertree parsing
		my ($equivFile, $treeFile);
		my (%STnodeList, $numSTclade, @STtaxa, %taxonST);
		
	# STR parameters
		my (@STRtaxa, %isSTR, %readIndexTaxa, %indexTaxa, %numIndices);
		my (%indexCount, @indexList);
		my @reversedTrees;
	
	# Global variables
		my (@AllTrees, %treeNames);
		my $block;
		my $debug = 0;
		my $verbose = 0;

for (my $i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i] =~ /-r(.+)/)
		{
		$equivFile = $1;
		}
	elsif ($ARGV[$i] =~ /-t(.+)/)
		{
		$treeFile = $1;
		}
	elsif ($ARGV[$i] =~ /-v/)
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] =~ /-x/)
		{
		$verbose = $debug = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: reverseSTR.pl -r<STR file> -r<treefile> [-v] [-h]\n";
		print "Options: -r<filename> = STRindexer.pl file listing taxonomic equivalents\n";
		print "         -t<treefile> = tree file containing gene trees in NEXUS format\n";
		print "         -v = verbose output\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "Usage: reverseSTR.pl -r<STR file> -r<treefile> [-v] [-h]\n";
		exit(1); 
		}
	}

if (not defined $treeFile or not defined $equivFile)
	{
	print "ERROR: Must specify file listing taxonomic equivalents.\n" if (not defined $equivFile);
	print "ERROR: Must specify nexus-formatted treefile.\n" if (not defined $treeFile);
	die;
	}

(my $outTreeFile = $treeFile) =~ s/\.\w*$//;
	$outTreeFile .= "_rvsSTR.tre";
	
# Read in and store taxonomic equivalents
	print "Reading in taxonomic equivalents from $equivFile ...\n";
	my $equivZero = time;
	
	setLineBreak($equivFile);
	open (EQUIV, "<$equivFile") or die "Cannot open file containg taxonomic equivalents, $equivFile\n";
		while (<EQUIV>)
			{
			chomp;
			next unless ($_);
			next if ($_ =~ /^Taxon/);
			last if ($_ =~ /^======/);
			my @indexTaxa = split(/\s+/, $_);
			
			my $STR = shift @indexTaxa;	# Get STR taxon and process
				push @STRtaxa, $STR;
				$isSTR{$STR} = 1;
				
			push @ {$readIndexTaxa{$STR} }, $_ foreach (@indexTaxa);	# Associate (multiple) index taxa with STR taxon
			}
	close EQUIV;
	
	# Check that no index taxa are themselves STRs
		foreach my $STR (@STRtaxa)
			{
			foreach my $index (@ {$readIndexTaxa{$STR} })
				{
				next if (defined $isSTR{$index});
				push @ {$indexTaxa{$STR} }, $index;
				$indexCount{$index}++;
					push @indexList, $index if ($indexCount{$index} == 1);
				}
			$numIndices{$STR} = scalar(@ {$indexTaxa{$STR} });
			}

		@STRtaxa = sort { $numIndices{$b} <=> $numIndices{$a} } keys %numIndices;
#			if ($debug)
#				{
#				foreach my $STR (@STRtaxa)
#					{
#					print "\t$STR ($numIndices{$STR}):";
#					print "\t$_" foreach (@ {$indexTaxa{$STR} });
#					print "\n";
#					}
#				}
	
	printf "\t%s equivalent taxa read in and processed in %s seconds\n", scalar(@STRtaxa), time - $equivZero;
	printf "\t%s unique index taxa present\n", scalar(@indexList);

# Read in supertree and process
	my $treeCount = 0;
	$block = 0;
	treeRead($treeFile);
	
	foreach my $superTree (@AllTrees)
		{
		$treeCount++;
		printf "Processing tree $treeCount of %s ($treeNames{$treeCount})\n", scalar(@AllTrees);
	
		# Tree-specific reverseSTR parameters
			my %STRlist;
			my @rejectList;

		$superTree =~ s/:\d+\.?\d*e?\-?\d*//g;	# Remove any branch length information
			$numSTclade = ($superTree =~ tr/(//);
				die "ERROR: Supertree does not have enough informative nodes ($numSTclade clades)\n" if ($numSTclade < 2);
			print "\nSupertree description: $superTree\n" if ($verbose);

		# Parse tree to get clade memberships and sizes
			my (%ParentClade, %CladeDescrip, %CladeSupport, %STcladeSize, %open, %start, %stop, %present);
			my ($position, $CladeID, $CurrentClade) = (0, 0, 0);
			my $OldCurrent;
			while ($position < length($superTree))
				{
				my $offset = 1;
				if (substr($superTree, $position, 1) eq "(")		#New clade; go one level deeper
					{
					$CladeID++;
					$ParentClade{$CladeID} = $CurrentClade;
					$CurrentClade = $CladeID;
					$open{$CurrentClade} = 1;
					$start{$CladeID} = $position;
					}
				elsif (substr($superTree, $position, 1) eq ")")	#Clade finished; clean up and go up one level
					{
					$open{$CurrentClade} = 0;
					$OldCurrent = $CurrentClade;
					$CurrentClade = $ParentClade{$OldCurrent};
					$stop{$OldCurrent} = $position;
					$CladeDescrip{$OldCurrent} = substr($superTree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
					$CladeDescrip{$OldCurrent} =~ s/:\d+\.?\d*e?\-?\d*//g;	# Remove branch length information
					}
				elsif (substr($superTree, $position, 1) eq ",")
					{
					# Do nothing
					}
				else	# Determine membership of clade
					{
					until (substr($superTree, $position+$offset, 1) eq "(" or substr($superTree, $position+$offset, 1) eq ")" or substr($superTree, $position+$offset, 1) eq "," or substr($superTree, $position+$offset, 1) eq ":")
						{
						$offset++;
						}
					my $taxon = substr($superTree, $position, $offset);
						push @STtaxa, $taxon;
						$taxonST{$taxon} = 1;
						$ParentClade{$taxon} = $CurrentClade;
	#				for (my $clade = 1; $clade <= $CladeID; $clade++)	# Assign all open clades to taxon and increment size of open clades
	#					{
	#					if ($open{$clade})
	#						{
	#						push @{ $STnodeList{$taxon} }, $clade;
	#						$STcladeSize{$clade}++;
	#						}
	#					}
					$position += $offset-1;
					}
				$position++;
				}
		
			@STtaxa = sort @STtaxa;
		
#			if ($verbose)
#				{
#				print "\nTaxon membership and size of supertree clades:\n";
#					for (my $clade = 1; $clade <= $numSTclade; $clade++)
#						{
#						print "\t$clade ($STcladeSize{$clade}): $CladeDescrip{$clade}\n" if ($verbose);
#						} 
#				print "\nSupertree clade membership of taxa:\n";
#					foreach my $taxon (@STtaxa)
#						{
#						print "\t$taxon:";
#						foreach my $node (@{ $STnodeList{$taxon} })
#							{
#							print "\t$node"
#							}
#						print "\n";
#						}
#				}

		# Add STR taxa to tree where possible
			print "\tAdding STR taxa to tree ...\n";
			my $reverseZero = time;
			foreach my $STR (@STRtaxa)
				{
				if ($verbose)
					{
					print "\t\tProcessing $STR ...\n";
					print "\t\t\tIndex taxa ($numIndices{$STR}):";
					print "\t$_" foreach (@ {$indexTaxa{$STR} });
					print "\n";
					}

				if (defined $taxonST{$STR})
					{
					print "\t\t\tAlready present on tree; skipping to next taxon\n" if ($verbose);
					next;
					}
				
				if ($numIndices{$STR} > 1)
					{
					print "\t\t\tChecking depth of multiple index taxa ...\n" if ($verbose);
					my $errorFlag = 0;
					my $parentNode = $ParentClade{$indexTaxa{$STR}[0]};	# Get parent clade of first index taxon as reference
					foreach my $index (@{ $indexTaxa{$STR} })
						{
						if ($ParentClade{$index} != $parentNode)
							{
							print "\t\t\t\tIndex taxon $index present in different clade (may be true for other taxa as well); $STR cannot be placed unambiguously\n" if ($verbose);
							push @rejectList, $STR;
							$errorFlag = 1;
							last;
							}
						}
						
					if ($errorFlag == 0)	# Replace first index taxon with "index, STR"
						{
						print "\t\t\t\tAll index taxa present in same clade; $STR will be added to same clade\n" if ($verbose);
						$superTree =~ s/$indexTaxa{$STR}[0],/$indexTaxa{$STR}[0], $STR,/;
						$superTree =~ s/$indexTaxa{$STR}[0]\)/$indexTaxa{$STR}[0], $STR\)/;
						}
					}
				else	# STR taxon has a single index taxon
					{
					push @{ $STRlist{$indexTaxa{$STR}[0]} }, $STR;	# Store all STRs for the singleton index taxon to be later added simultaneously
					print "\t\t\tWill be added to index taxon $indexTaxa{$STR}[0]\n" if ($verbose);
					}
				}
		
			# Add in all STRs for singleton index taxa
				print "\t\tProcessing singleton index taxa ...\n";
				foreach my $index (@indexList)
					{
					next unless (defined @{ $STRlist{$index} });
					print "\t\t\tProcessing $index ...\n" if ($verbose);
					my $replaceList = join(", ", @{ $STRlist{$index} });
					print "\t\t\t\tSTR taxa: $replaceList\n" if ($verbose);
					
					$superTree =~ s/$index,/\($index, $replaceList\),/;
					$superTree =~ s/$index\)/\($index, $replaceList\)\)/;
					}
				
			# Adjust branch lengths in tree according to status of taxon
				$superTree =~ s/\)/\):1/g;	# Give all nodes a branch length of 1
				
				foreach my $taxon (@STtaxa)	# All normal taxa given a branch length of 1
					{
					$superTree =~ s/$taxon,/$taxon:1,/;
					$superTree =~ s/$taxon\)/$taxon:1\)/;
					}
					
				foreach my $STR (@STRtaxa)	# STR taxa given a branch length of 0
					{
					next if (defined $taxonST{$STR});
					$superTree =~ s/$STR,/$STR:0,/;
					$superTree =~ s/$STR\)/$STR:0\)/;
					}
					
				push @reversedTrees, $superTree;
		
			if (@rejectList)
				{
				printf "\n\t\tThe following %s STR taxa could not be placed unambiguously:\n", scalar(@rejectList);
				print "\t\t\t$_\n" foreach (@rejectList);
				}
			else
				{
				print "\n\tAll STR taxa placed unambiguously\n";
				}
			printf "\n\t\tTime taken: %s seconds\n", time - $reverseZero;
		}
	
# Print tree(s)!
	$treeCount = 0;
	open (TREE, ">$outTreeFile") or die "Could not write tree to file $outTreeFile\n";
		print TREE "#NEXUS\n\nbegin trees;\n";
		foreach my $tree (@reversedTrees)
			{
			$treeCount++;
			print TREE "\tTREE $treeNames{$treeCount} = $tree;\n";
			}
		print TREE "end;\n";
	close TREE;

exit(0);

### Subroutines used in script

sub treeRead		# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
	{
	my $treeFile = shift @_;
	undef @AllTrees;
	setLineBreak($treeFile);
	open (fTREES, "<$treeFile") or die "Cannot open treeFile $treeFile\n";
		my ($treeDescrip, $treeFlag, $treeCount);
		my (@taxonList, %taxonPresent);
		my $treeBlock = 0;
		my (%TTable, $translation, $translateFlag, $maxTaxonID, $transText);
		
		while (<fTREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			if ($_ =~ m/begin trees;/i)		# Clear variables for each tree block
				{
				undef %TTable;
				$translation = $translateFlag = $treeFlag = $maxTaxonID = $treeCount = 0;
				$treeBlock++;
				}
			if ($_ =~ m/translate/i)	#Read translation table of taxon names (if any)
				{
				$translateFlag = $translation = 1;
				$transText = "";
				print "Reading translation table ...\n" if ($verbose);
				}
			if ($translateFlag)
				{
				unless ($_ =~ m/translate/i)
					{
					$transText .= $_;
					if ($_ =~ m/;/)
						{
						$translateFlag = 0;
						$transText =~ s/;$//;	#Strip trailing semi-colon
						my @transEntries = split(/,/, $transText);
						foreach my $entry (@transEntries)
							{
							$entry =~ s/^\s+//;		#Strip (any) leading white space
							my ($taxonID, $taxonName) = split(/\s/,$entry);
							$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
							$taxonName =~ s/\s+//g;
							$TTable{$taxonID} = $taxonName;
							}
						}
					}
				}
			if ($_ =~ /tree/i and $_ =~ /=/)
				{
				unless ($_ =~ /^>/)		#Ignore (any) commented lines
					{
					$treeDescrip = "";
					unless ($block == 0 and $treeBlock > 1)
						{
						$treeFlag = 1;
						}
					}
				}
			if ($treeFlag)
				{
				$treeDescrip .= $_;
				if ($_ =~ m/;/)
					{
					$treeCount++;
					print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
					print "\tTranslating taxon names ...\n" if ($verbose and $translation);
					$treeFlag = 0;
					my $position = 0;
					until (substr($treeDescrip, $position, 1) eq "(")		#Avoid anything before tree description proper
						{
						$position++;
						}
					my $tree = substr($treeDescrip, $position);
						$tree =~ s/;\s*$//;
						$tree =~ s/\s+//g;
					my ($name) = (split /=/, $treeDescrip, 0);		#Get name of tree, if any
						$name =~ s/\s*\w*tree\s+//i;
						$name =~ s/\s*\*\s*//i;
						$name =~ s/\s+$//g;
						$treeNames{$treeCount} = $name;
					my $openCount = ($tree =~ tr/(//);
					my $closedCount = ($tree =~ tr/)//);
					if ($openCount == $closedCount)		#Tree is balanced; process
						{
						$position = length($tree) - 1;
						while ($position >= 0)
							{
							my $IDlength = 1;
							if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
								{
								$position--;
								}
							else	# Read in taxonID
								{
								until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
									{
									$position--;
									$IDlength++;
									}
								unless (substr($tree, $position, 1) eq ":")		# Avoid reading branch lengths as taxon names
									{
									my $readName = substr($tree, $position+1, $IDlength-1);
									if ($translation)	# Translate taxon names
										{
										if ($TTable{$readName})
											{
											substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		#Translate name
											unless ($taxonPresent{$TTable{$readName}})
												{
												push @taxonList, $TTable{$readName};
												$taxonPresent{$TTable{$readName}} = 1;
												}
											}
										else
											{
											print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
											substr($tree, $position+1, $IDlength-1) = "xxx";		#Translate name
											unless ($taxonPresent{"xxx"})
												{
												push @taxonList, "xxx";
												$taxonPresent{"xxx"} = 1;
												}
											}
										}
									else	#Enter name as read from tree, unless already present
										{
										unless ($taxonPresent{$readName})
											{
											push @taxonList, $readName;
											$taxonPresent{$readName} = 1;
											}
										}
									}
								}
							}
						$tree =~s /\)\w+/\)/g;	# Remove any named nodes
						push @AllTrees, $tree;
					print "\tTree description:\n\t\t$tree\n" if ($verbose);
						}
					else	#Tree is unbalanced; reject
						{
						print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
						}
					}
				}
			}
	close fTREES;

	return (@AllTrees, %treeNames);
	}

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

# Version history
#
#	v1.0 (January 26, 2005)
#		- initial release
